/******************************************************************************
 * fileman.cpp - Manage reading files.
 * Note file reading is multithreaded
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QThread>
#include <QMutex>

#include "world.h"
#include "nbtfile.h"
#include "nbtobject.h"
#include "regionfile.h"
#include "resourcedock.h"
#include "mainwindow.h"
#include "fileman.h"

static const bool DEBUG = false;
const int maxRegionsInMemory = 8;

QMutex workerCountMutex;
QMutex fileQueueMutex;

FileMan::FileMan(MainWindow* parent)
	: m_mainWindow(parent),
	  m_workerCount(0),
	  m_regionsLoaded(0)
{
}

void FileMan::setWorld(World *world)
{
	m_world = world;
	m_regionsLoaded = 0;
}

bool FileMan::queuedFile(NBTFile *file)
{
	fileQueueMutex.lock();
	bool res = m_fileQueue.contains(file);
	fileQueueMutex.unlock();
	return(res);
}

void FileMan::readFile(NBTFile* file)
{
	if (file->isLoaded()) {
		file->setLastUse();
		emit readFileDone(file);
		return;
	}
	fileQueueMutex.lock();
	bool res = m_fileQueue.contains(file);
	fileQueueMutex.unlock();
	if (res)
		return;
	RegionFile* rf = dynamic_cast<RegionFile*>(file);
	if (rf) {
		if (DEBUG)
			qDebug() << "FileMan:readFile queue" << rf->getFileName();
		if (m_regionsLoaded >= maxRegionsInMemory) {
			QDateTime dt;
			Group* regions = m_world->getGroup(wgOverworldRegions);
			RegionFile* oldest = nullptr;
			foreach (NBTFile* f, regions->getFiles()) {
				RegionFile* rf = (RegionFile*)f;
				if (rf->m_loaded) {
					if (!dt.isValid()) {
						dt = rf->m_lastUse;
						oldest = rf;
					} else {
						if (rf->m_lastUse < dt) {
							dt = rf->m_lastUse;
							oldest = rf;
						}
					}
				}
			}
			if (oldest) {
				oldest->unload();		// XXX need to respect the dirty flag
				m_regionsLoaded--;
			}
		}
		m_regionsLoaded++;
	}
	fileQueueMutex.lock();
	file->m_fileManReading = false;
	m_fileQueue.append(file);
	fileQueueMutex.unlock();

	if (m_workerCount < numParallelReads) {
		NBTFile* toReadFile = selectFileForReading();
		if (!toReadFile) {
			qWarning() << "FileMan: I just put a file on the queue and it's not there. Maybe this is ok." << file->getFileName();
			return;
		}
		threadReadStart(toReadFile);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////
/// @brief Determine the next file queued to be read.
/// Note this runs in the FileReaderThread.
///
/// @return The file that should be read. This file is marked as being read, so no "just peeking".
///
NBTFile* FileMan::selectFileForReading()
{
	fileQueueMutex.lock();
	NBTFile* file = nullptr;
	foreach (NBTFile* f, m_fileQueue) {
		if (!f->m_fileManReading) {
			file = f;
			break;
		}
	}
	if (file)
		file->m_fileManReading = true;
	fileQueueMutex.unlock();
	return(file);
}

void FileMan::threadReadStart(NBTFile* file)
{
	workerCountMutex.lock();
	m_workerCount++;
	workerCountMutex.unlock();
	if (DEBUG)
		qDebug() << "FileMan:threadReadStart" << file->getFileName();
	FileReaderThread* frt = new FileReaderThread();
	connect(frt, SIGNAL(readFileDone(NBTFile*)), this, SLOT(threadReadFileDone(NBTFile*)));
	connect(frt, &FileReaderThread::finished, frt, &QObject::deleteLater);
	frt->setupRead(file, this);
	frt->start();
}

void FileMan::threadReadFileDone(NBTFile* file)
{
	workerCountMutex.lock();
	m_workerCount--;
	workerCountMutex.unlock();
	fileQueueMutex.lock();
	m_fileQueue.removeOne(file);
	file->m_fileManReading = false;
	fileQueueMutex.unlock();
	emit (readFileDone(file));
	NBTFile* nextFile = selectFileForReading();
	if (nextFile)
		threadReadStart(nextFile);
}

void FileMan::errorMsgHandler(const QString& msg)
{
	emit(sendErrorMsg(msg));
}

///////////////////////////////////////////////////////////////////////////////////////////////////
/// \brief RailMan is done processing. Clean up any region files loaded (release 3.5GB of memory)
/// Don't unload a RegionFile if the ResourceDock is looking at it.
///
void FileMan::onRailManDone()
{
	const NBTFile* resourceDockFile = m_mainWindow->getResourceDock()->getCurrentFile();
	fileQueueMutex.lock();
	for (int i=m_fileQueue.size()-1; i>=0; i--) {
		NBTFile* file = m_fileQueue.at(i);
		RegionFile* rf = dynamic_cast<RegionFile*>(file);
		if (rf) {
			if (rf == resourceDockFile)
				continue;
			if (!rf->m_fileManReading) {
				m_fileQueue.removeOne(rf);
				m_regionsLoaded--;
			}
		}
	}
	fileQueueMutex.unlock();
	Group* regions = m_world->getGroup(wgOverworldRegions);
	foreach(NBTFile* file, regions->getFiles()) {
		RegionFile* rf = dynamic_cast<RegionFile*>(file);
		if (rf) {
			if (rf == resourceDockFile)
				continue;
			if (rf->isLoaded()) {
				rf->unload();
			}
		}
	}
}

//=================================================================================================
void FileReaderThread::setupRead(NBTFile* file, FileMan* parent)
{
	m_file = file;
	m_fileMan = parent;
}

void FileReaderThread::run()
{
	m_file->load(m_fileMan);
	emit(readFileDone(m_file));
	//this->deleteLater();
}

