/******************************************************************************
 * maintoolbar.cpp - Main Toolbar
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QMenuBar>
#include <QDoubleSpinBox>

#include "mainwindow.h"
#include "commandman.h"
#include "consoleman.h"
#include "helpman.h"
#include "syncman.h"
#include "mapwidget.h"
#include "maintoolbar.h"

#define iString(name) QString(":/resources/icons/").append(name).append(sizeString)

MainToolbar::MainToolbar(int iconSize, MainWindow* parent)
	: QToolBar(parent),
	  m_mainWindow(parent)

{
	QString title = "MainToolbar";
	setWindowTitle(title);
	setObjectName(title);
	QString sizeString;
	sizeString = QString("_%1x%2.png").arg(iconSize).arg(iconSize);

	m_openAction = addAction(QIcon(iString("Open")), "Open World",
						   m_mainWindow, SLOT(openWorld()));
	m_openAction->setShortcut(QKeySequence::Open);
	m_saveAction = addAction(QIcon(iString("Save")), "Save World", m_mainWindow, SLOT(saveWorld()));
	m_saveAction->setShortcut(QKeySequence::Save);
	//m_saveAction->setDisabled(true);
	addSeparator();
	CommandMan* commandMan = m_mainWindow->getCommandMan();
	m_undoAction = addAction(QIcon(iString("Undo")), "Undo", commandMan, SLOT(undo()));
	m_undoAction->setShortcut(QKeySequence::Undo);
	m_redoAction = addAction(QIcon(iString("Redo")), "Redo", commandMan, SLOT(redo()));
	m_redoAction->setShortcut(QKeySequence::Redo);

	addSeparator();
	m_zoomSpinBox = new QSpinBox();
	m_zoomSpinBox->setMaximum(1000);
	m_zoomSpinBox->setValue(100);
	//m_zoomSpinBox->setMinimum(10);
	m_zoomSpinBox->setSuffix("%");
	//m_zoomSpinBox->setSingleStep(0.5);
	setZoomValue(1.0);
	connect(m_zoomSpinBox, SIGNAL(valueChanged(int)), this, SLOT(onZoomValueChanged(int)));
	this->addWidget(m_zoomSpinBox);

	SyncMan* syncMan = m_mainWindow->getSyncMan();
	m_syncAction = addAction(QIcon(iString("Synchronize")), "Synchronize", syncMan, SLOT(onSynchronize(bool)));
	m_syncAction->setCheckable(true);

	m_prefsAction = addAction(QIcon(iString("Settings")), "Preferences", m_mainWindow, SLOT(onPreferences()));
	setupMenu(sizeString);
	commandMan->setUndoRedo(m_undoAction, m_redoAction, m_saveAction, m_syncAction);
}

void MainToolbar::setupMenu(const QString &sizeString)
{
	QMenu* menu = m_mainWindow->menuBar()->addMenu(tr("&File"));
	menu->addAction(m_openAction);
	menu->addAction(m_saveAction);
	menu->addSeparator();
	menu->addAction("Export PNG", m_mainWindow, SLOT(exportPNG()));
	menu->addAction(tr("Quit"), m_mainWindow, SLOT(close()))->setShortcut(QKeySequence::Quit);

	menu = m_mainWindow->menuBar()->addMenu(tr("&Edit"));
	menu->addAction(m_undoAction);
	menu->addAction(m_redoAction);
	menu = m_mainWindow->menuBar()->addMenu(tr("&Tools"));
	menu->addAction(QIcon(iString("map")), tr("&Map Settings"), m_mainWindow->getMapWidget(), SLOT(onMapSettingsDialog(bool)));
	menu->addAction(QIcon(iString("rail")), tr("&Railman"), m_mainWindow->getMapWidget(), SLOT(onRailManDialog(bool)));
	menu->addAction(QIcon(iString("console")), tr("&console"), m_mainWindow->getConsoleMan(), SLOT(onConsoleDialog(bool)));
	menu->addSeparator();
	menu->addAction(QIcon(iString("Synchronize")), tr("&Live Updates"), m_mainWindow->getSyncMan(), SLOT(onSynchronize(bool)));
	menu->addSeparator();
	menu->addAction(QIcon(iString("Settings")), tr("Preferences"), m_mainWindow, SLOT(onPreferences()));

	menu = m_mainWindow->menuBar()->addMenu(tr("&Help"));
	menu->addAction(QIcon(iString("Help")), tr("&Help"), m_mainWindow->getHelpMan(), SLOT(onHelp()));
	menu->addAction(QIcon(iString("Information")), tr("&About"), m_mainWindow->getHelpMan(), SLOT(onAboutDialog()));
}

void MainToolbar::setZoomValue(double zoom)
{
	int val = zoom*100;
	m_zoomSpinBox->setValue(val);
	m_zoomSpinBox->setSingleStep(val/20);	// 5%
}

void MainToolbar::onZoomValueChanged(int value)
{
	double d = value/100.0;
	setZoomValue(d);
	emit(zoomValueChanged(d));
}

void MainToolbar::resetWidgets()
{
	m_syncAction->setChecked(false);
}

