/******************************************************************************
 * nbtobject.cpp - A single NBT object
 * See: https://minecraft.gamepedia.com/NBT_format
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>

#include "nbtobject.h"

QStringList	nbtTypeNames;

void NBTObject::init()
{
	nbtTypeNames.clear();
	nbtTypeNames << "" << "Byte" << "Short" << "Int" << "Long" << "Float" << "Double" << "Byte Array"
				 << "String" << "List" << "Compound" << "Int Array" << "Long Array" ;
}


NBTObject::~NBTObject()
{
	switch (type) {
	case TAG_Byte_Array:
		delete d_byte_array;
		break;
	case TAG_String:
		delete d_string;
		break;
	case TAG_Compound:
		if (d_compound) {
			while (!d_compound->isEmpty())
				delete d_compound->takeFirst();
			delete d_compound;
		}
		break;
	case TAG_List:
		delete d_list;
		break;
	case TAG_Int_Array:
		delete[] d_int_array;
		break;
	case TAG_Long_Array:
		delete[] d_long_array;
		break;
	case TAG_End:
	case TAG_Byte:
	case TAG_Short:
	case TAG_Int:
	case TAG_Long:
	case TAG_Float:
	case TAG_Double:
		break;
	default:
		qCritical() << "~NBTObject unhandled type" << type;
		break;
	}
}

NBTObject* NBTObject::findChild(const QString& name)
{
	QStringList segments = name.split(":");
	if (this->name == segments[0] || segments[0] == "*") {
		segments.takeFirst();
		if (segments.isEmpty())
			return(this);
		if (this->type != TAG_Compound)
			return(nullptr);
		QString cname = segments.join(":");
		foreach (NBTObject* child, *this->d_compound) {
			NBTObject* res = child->findChild(cname);
			if (res)
				return(res);
		}
	}
	return(nullptr);
}


NBTObject* NBTObject::removeChild(int row)
{
	NBTObject* child = nullptr;
	if (this->type == TAG_Compound) {
		child = d_compound->at(row);
		d_compound->removeAt(row);
	} else if (this->type == TAG_List) {
		child = d_list->removeRow(row);
	}
	return(child);
}

void NBTObject::insertChild(int row, NBTObject *child)
{
	if (this->type == TAG_Compound) {
		d_compound->insert(row, child);
	} else if (this->type == TAG_List) {
		d_list->insertRow(row, child);
	}
}

int NBTObject::getChildIndex(void* child)
{
	int i = -1;
	if (type == TAG_List) {
		NList* nlist = d_list;
		switch (nlist->objtype) {
		case TAG_Compound:
			for (i=0; i<nlist->objcount; i++) {
				if (nlist->d_compound->at(i) == child)
					return(i);
			}
			return(-1);
		case TAG_List:
			for (i=0; i<nlist->objcount; i++) {
				if (nlist->d_list->at(i) == child)
					return(i);
			}
			return(-1);
		}
	}
	return(i);
}

void NBTObject::setValue(QVariant value)
{
	switch(type) {
	case TAG_Byte:
		d_byte = value.toInt();
		break;
	case TAG_Short:
		d_short = value.toInt();
		break;
	case TAG_Int:
		d_int = value.toInt();
		break;
	case TAG_Long:
		d_long = value.toLongLong();
		break;
	case TAG_Float:
		d_float = value.toFloat();
		break;
	case TAG_Double:
		d_double = value.toDouble();
		break;
	case TAG_String:
		d_string = new QString(value.toString());
		break;
	default:
		qWarning() << "NBTObject:setValue type" << nbtTypeNames[type] << "unsupported";
	}
}

QVariant NBTObject::getValue()
{
	switch(type) {
	case TAG_Byte:
		return(QVariant(d_byte));
	case TAG_Short:
		return(QVariant(d_short));
	case TAG_Int:
		return(QVariant(d_int));
	case TAG_Long:
		return(QVariant(d_long));
	case TAG_Float:
		return(QVariant(d_float));
	case TAG_Double:
		return(QVariant(d_double));
	case TAG_String:
		return(QVariant(*d_string));
		break;
	default:
		qWarning() << "NBTObject:getValue type" << nbtTypeNames[type] << "unsupported";
	}
	return(QVariant());
}

NBTObject& NBTObject::operator=(const NBTObject& other)
{
	this->type = other.type;
	switch(this->type) {
	case TAG_Byte:
		d_byte = other.d_byte;
		break;
	case TAG_Short:
		d_short = other.d_short;
		break;
	case TAG_Int:
		d_int = other.d_int;
		break;
	case TAG_Long:
		d_long = other.d_long;
		break;
	case TAG_Float:
		d_float = other.d_float;
		break;
	case TAG_Double:
		d_double = other.d_double;
		break;
	case TAG_String:
		d_string = new QString(*other.d_string);
		break;
	default:
		qWarning() << "NBTObject:operator= type" << nbtTypeNames << "unsupported";
	}
	return(*this);
}
