/******************************************************************************
 * preferencesdialog.cpp - Dialog to control various settings
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2019 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QGridLayout>
#include <QLabel>
#include <QPushButton>
#include <QLineEdit>
#include <QSpinBox>
#include <QCheckBox>
#include <QFileDialog>

#include "world.h"
#include "yamcedoc.h"
#include "helpman.h"
#include "consoleman.h"
#include "mainwindow.h"
#include "preferencesdialog.h"


PreferencesDialog::PreferencesDialog(MainWindow* mainWindow)
	: QDialog(mainWindow),
	  m_mainWindow(mainWindow)
{
	m_world = m_mainWindow->getWorld();
	QVBoxLayout* layout = new QVBoxLayout();
	QTabWidget* tabWidget = new QTabWidget(this);
	m_worldTab = new WorldTab(this);
	QString s = "N/A";
	if (m_world)
		s = m_world->getName();
	tabWidget->addTab(m_worldTab, s);
	if (!m_world)
		m_worldTab->setEnabled(false);

	m_systemTab = new SystemTab(this);
	tabWidget->addTab(m_systemTab, tr("System"));

	this->setWindowTitle("yamce - Preferences");
	layout->addWidget(tabWidget);
	this->setLayout(layout);
}

void PreferencesDialog::setWorld(World* world)
{
	m_world = world;
	m_worldTab->setWorld(m_world);
	//m_systemTab->setWorld(m_world);
}

SystemTab::SystemTab(PreferencesDialog* parent)
{
	m_mainWindow = parent->getMainWindow();

	QVBoxLayout*	vlayout = new QVBoxLayout();
	QGridLayout*	glayout = new QGridLayout();
	QLabel* l = new QLabel(tr("Help Files Directory:"));
	glayout->addWidget(l, 0, 0);

	m_helpDirLineEdit = new QLineEdit();
	m_helpDirLineEdit->setText(m_mainWindow->getHelpMan()->getHelpDirectory());
	connect(m_helpDirLineEdit, SIGNAL(editingFinished()), this, SLOT(onHelpDirectoryChanged()));
	glayout->addWidget(m_helpDirLineEdit, 0, 1);

	QPushButton* browseButton = new QPushButton();
	browseButton->setText(tr("Browse"));
	browseButton->setDefault(false);
	browseButton->setAutoDefault(false);
	connect(browseButton, SIGNAL(clicked(bool)), this, SLOT(onHelpDirectoryBrowse()));
	glayout->addWidget(browseButton, 0, 2);

	m_showDebug = new QCheckBox("Show debug in console");
	m_showDebug->setCheckState(m_mainWindow->getConsoleMan()->getShowDebug() ? Qt::Checked : Qt::Unchecked);
	connect(m_showDebug, SIGNAL(stateChanged(int)), this, SLOT(onShowDebugToggled(int)));
	glayout->addWidget(m_showDebug, 1, 0, 1, 3);

	vlayout->addLayout(glayout);
	vlayout->addStretch(2);
	this->setLayout(vlayout);
}

void SystemTab::onHelpDirectoryChanged()
{
	m_mainWindow->getHelpMan()->setHelpDirectory(m_helpDirLineEdit->text());
}

void SystemTab::onHelpDirectoryBrowse()
{
	QString dir = QFileDialog::getExistingDirectory(this, tr("Select Directory"));
	if (!dir.isEmpty()) {
		m_helpDirLineEdit->setText(dir);
		m_mainWindow->getHelpMan()->setHelpDirectory(m_helpDirLineEdit->text());
	}
}

void SystemTab::onShowDebugToggled(int newState)
{
	m_mainWindow->getConsoleMan()->setShowDebug(newState);
}

WorldTab::WorldTab(PreferencesDialog* parent)
	: m_world(nullptr)
{
	QVBoxLayout*	vlayout = new QVBoxLayout();
	QGridLayout*	glayout = new QGridLayout();
	QLabel* l = new QLabel(tr("MapWall dirname:"));
	glayout->addWidget(l, 0, 0);

	m_mapWallLineEdit = new QLineEdit();
	//m_mapWallLineEdit->setText(m_mainWindow->getWorld()->m_yamceDoc.m_exportPngDir);
	connect(m_mapWallLineEdit, SIGNAL(editingFinished()), this, SLOT(onMapWallEdited()));
	glayout->addWidget(m_mapWallLineEdit, 0, 1);

	QPushButton* browseButton = new QPushButton();
	browseButton->setText(tr("Browse"));
	browseButton->setDefault(false);
	browseButton->setAutoDefault(false);
	connect(browseButton, SIGNAL(clicked(bool)), this, SLOT(onMapWallBrowse(bool)));
	glayout->addWidget(browseButton, 0, 2);

	l = new QLabel("# of maps to watch:");
	glayout->addWidget(l, 1, 0);
	m_mapsToWatch = new QSpinBox();
	m_mapsToWatch->setRange(0, 1000);
	//m_mapsToWatch->setValue(m_mainWindow->getWorld()->m_yamceDoc.m_maxMapsWatched);
	connect(m_mapsToWatch, SIGNAL(valueChanged(int)), this, SLOT(onMapsToWatchChanged(int)));
	glayout->addWidget(m_mapsToWatch, 1, 1, 1, 1);
	vlayout->addLayout(glayout);
	vlayout->addStretch(2);
	this->setLayout(vlayout);
	if (parent->getWorld())
		setWorld(parent->getWorld());
}

void WorldTab::setWorld(World* world)
{
	m_world = world;
	m_mapWallLineEdit->setText(m_world->m_yamceDoc.m_exportPngDir);
	m_mapsToWatch->setValue(m_world->m_yamceDoc.m_maxMapsWatched);

}
void WorldTab::onMapWallEdited()
{
	m_world->m_yamceDoc.m_exportPngDir = m_mapWallLineEdit->text();
}

void WorldTab::onMapWallBrowse(bool)
{
	QString s = QFileDialog::getExistingDirectory(this, tr("Choose name of MapWall PNG dir"));
	if (s.isEmpty()) {
		qInfo() << "MapWall filename cancelled";
		return;
	}
	m_world->m_yamceDoc.m_exportPngDir = s;
	m_mapWallLineEdit->setText(s);
}

void WorldTab::onMapsToWatchChanged(int value)
{
	m_world->m_yamceDoc.m_maxMapsWatched = value;
}
