/******************************************************************************
 * railmandialog.cpp - Dialog to control RailMan
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2019 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QGroupBox>
#include <QGridLayout>
#include <QCheckBox>
#include <QLabel>
#include <QTextEdit>
#include <QPushButton>
#include <QTimer>
#include <QDialog>

#include "fileman.h"
#include "mapfile.h"
#include "mainwindow.h"
#include "mapcolordialog.h"
#include "nbtobject.h"
#include "regionfile.h"
#include "railman.h"
#include "mapwidget.h"
#include "world.h"
#include "yamcedoc.h"
#include "railmandialog.h"

RailManDialog::RailManDialog(MapWidget* parent, MainWindow* mainWindow)
	: QDialog(parent),
	  m_mainWindow(mainWindow),
	  m_mapColorDialog(nullptr),
	  m_yamceDoc(nullptr),
	  m_railManIsRunning(false),
	  m_timer(nullptr),
	  m_isClosing(false)

{
	m_railMan = m_mainWindow->getRailMan();
	const int numCols=3;
	const int row=0;
	const int rowSpan=1;
	enum {
		mainCol,
		widgCol,
		statCol
	};

	QGridLayout* layout = new QGridLayout();
	layout->setColumnMinimumWidth(0, 20);
	setWindowTitle("yamce - Railman");


	m_drawRailLines = new QCheckBox("Draw Rail Lines");
	connect(m_drawRailLines, SIGNAL(toggled(bool)), this, SLOT(drawRailLinesToggled(bool)));
	layout->addWidget(m_drawRailLines,			row+3, mainCol, rowSpan, numCols);

	QGroupBox* groupBox = new QGroupBox();
	m_railStartEditLayout = new QGridLayout();
	groupBox->setLayout(m_railStartEditLayout);
	groupBox->setTitle("Rail Search Start Points");
	layout->addWidget(groupBox,					row+4, widgCol, rowSpan, numCols-1);

	setupRailStartEdits();

	m_railStatsEdit = new QTextEdit();
	layout->addWidget(m_railStatsEdit,			row+5, widgCol, rowSpan, numCols-1);
	layout->setRowStretch(row+6, 2);
	m_railStatsEdit->setHtml(m_railMan->getTextStats());

	m_colorButton = new QPushButton(" Rail Color");
	m_colorButton->setAutoDefault(false);
	connect(m_colorButton, SIGNAL(clicked(bool)), this, SLOT(onColorButtonClicked()));
	setColorButtonIcon();
	layout->addWidget(m_colorButton,			row+6, widgCol, rowSpan, numCols-1);

	m_processRailLines = new QPushButton("Scan for Rail Lines");
	m_processRailLines->setAutoDefault(false);
	connect(m_processRailLines, SIGNAL(clicked(bool)), this, SLOT(processRailLinesToggled(bool)));
	layout->addWidget(m_processRailLines,		row+7, widgCol, rowSpan, numCols-1);

	m_updateMinecraftMaps = new QPushButton("Update Minecraft maps");
	m_updateMinecraftMaps->setAutoDefault(false);
	connect(m_updateMinecraftMaps, SIGNAL(clicked(bool)), this, SLOT(onMinecraftMapsToggled(bool)));
	layout->addWidget(m_updateMinecraftMaps,	row+8, widgCol, rowSpan, numCols-1);

	this->setLayout(layout);
	maybeEnableOptions();
}

RailManDialog::~RailManDialog()
{
	while (!m_railStartEditList.isEmpty())
		delete m_railStartEditList.takeLast();	// Qt owns the widgets, but we own the container class
}

void RailManDialog::setWorld(World* world)
{
	m_world = world;
	m_yamceDoc = &world->m_yamceDoc;
	MapSettings& ms = m_yamceDoc->m_mapSettings;
	m_drawRailLines->setChecked(ms.m_drawRailLines);
	setupRailStartEdits();
	maybeEnableOptions();
	setColorButtonIcon();
}

void RailManDialog::setupRailStartEdits()
{
//	int count = 0;
	QList<XYZ*>*	startPoints = nullptr;
	if (m_yamceDoc) {
		startPoints = &m_yamceDoc->m_mapSettings.m_railStartPoints;
//		count = startPoints->size();
		int i;
		for (i=0; i<startPoints->size(); i++) {
			if (m_railStartEditList.size() <= i) {
				RailStartEditWithStatus* rsews = new RailStartEditWithStatus();
				rsews->m_lineEdit = new MSDLineEdit();
				connect(rsews->m_lineEdit, SIGNAL(editingFinished()), this, SLOT(railStartEditingFinished()));
				m_railStartEditLayout->addWidget(rsews->m_lineEdit,	i, 0);
				rsews->m_xyz = startPoints->at(i);
				XYZ* rp = startPoints->at(i);
				QString s = rp->toString(true);
				rsews->m_lineEdit->setText(s);
				rsews->m_status = new QLabel();
				rsews->m_railRootNeedsValidating = true;
				m_railStartEditLayout->addWidget(rsews->m_status,	i, 1);
				m_railStartEditList.append(rsews);
			} else {
				XYZ* rp = startPoints->at(i);
				QString s = rp->toString(true);
				RailStartEditWithStatus* rsews = m_railStartEditList.at(i);
				rsews->m_lineEdit->setText(s);
				rsews->m_lineEdit->setVisible(true);
				rsews->m_status->setVisible(true);
			}
		}
//		i++;
		if (m_railStartEditList.size() <= i) {
			RailStartEditWithStatus* rsews = new RailStartEditWithStatus();
			rsews->m_lineEdit = new MSDLineEdit();
			connect(rsews->m_lineEdit, SIGNAL(editingFinished()), this, SLOT(railStartEditingFinished()));
			m_railStartEditLayout->addWidget(rsews->m_lineEdit,	i, 0);
			rsews->m_status = new QLabel();
			m_railStartEditLayout->addWidget(rsews->m_status,	i, 1);
			m_railStartEditList.append(rsews);
		} else {
			RailStartEditWithStatus* rsews = m_railStartEditList.at(i);
			rsews->m_lineEdit->setVisible(true);
			rsews->m_lineEdit->clear();
			rsews->m_status->setVisible(true);
			rsews->m_status->setPixmap(QPixmap());
		}
		i++;
		for (; i<m_railStartEditList.size(); i++) {
			RailStartEditWithStatus* rsews = m_railStartEditList.at(i);
			rsews->m_lineEdit->setVisible(false);
			rsews->m_status->setVisible(false);

		}
	}
	updateRailRootStatus();
}

void RailManDialog::setColorButtonIcon()
{
	QPixmap* qpm = new QPixmap(40,40);
	int cindex = 18;		// default is red
	if (m_yamceDoc)
		cindex = m_yamceDoc->m_railColor;
	QColor color = getMapColorMap().at(cindex);
	qpm->fill(color);
	m_colorButton->setIcon(QIcon(*qpm));
	delete qpm;
}

void RailManDialog::maybeEnableOptions()
{
	bool b = m_drawRailLines->isChecked();
	m_updateMinecraftMaps->setEnabled(b);
	if (m_railManIsRunning) {
		m_processRailLines->setEnabled(false);
		m_updateMinecraftMaps->setEnabled(false);
	} else {
		bool c = b;
		if (c && m_yamceDoc->m_mapSettings.m_railStartPoints.isEmpty())
			c = false;
		m_processRailLines->setEnabled(c);
		c = b;
		if (m_railMan->getRails().isEmpty())
			c = false;
		m_updateMinecraftMaps->setEnabled(c);
	}
}

void RailManDialog::drawRailLinesToggled(bool b)
{
	m_yamceDoc->m_mapSettings.m_drawRailLines = b;
	maybeEnableOptions();
}

void RailManDialog::processRailLinesToggled(bool)
{
	m_railManIsRunning = true;
	maybeEnableOptions();
	emit(processRailsChecked());
	m_railStatsEdit->setHtml(m_railMan->getTextStats());
	m_timer = new QTimer();
	connect(m_timer, SIGNAL(timeout()), this, SLOT(onTimerTick()));
	m_timer->start(500);
}

void RailManDialog::onMinecraftMapsToggled(bool)
{
	m_railManIsRunning = true;
	maybeEnableOptions();
	m_railMan->onUpdateMinecraftMaps();
}

void RailManDialog::onTimerTick()
{
	m_railStatsEdit->setHtml(m_railMan->getTextStats());
}

static QString s_ErrorStyle = "background: #FFAAAA";
void RailManDialog::railStartEditingFinished()
{
	if (m_isClosing)
		return;
	QObject* object = sender();
	QLineEdit* qle = qobject_cast<QLineEdit*>(object);
	if (!qle) {
		qCritical() << "MapSettingsDialog:railStartEditingFinished unknown object called us";
		return;
	}
	RailStartEditWithStatus* rsews = findEditStatus(qle);
	if (!rsews) {
		qCritical() << "MapSettingsDialog:railStartEditingFinished can't find LineEdit";
		return;
	}
	QString let = rsews->m_lineEdit->text();
	if (let.isEmpty()) {
		if (rsews->m_xyz) {
//			if (m_railStartEditList.size() > 1) {
//				m_railStartEditLayout->removeWidget(rsews->m_lineEdit);
//				delete rsews->m_lineEdit;
//				m_railStartEditLayout->removeWidget(rsews->m_status);
//				delete rsews->m_status;
//			}
			//qDebug() << "RailManDialog:railStartEditingFinished rowcount =" << m_railStartEditLayout->rowCount();
			m_yamceDoc->m_mapSettings.m_railStartPoints.removeOne(rsews->m_xyz);
			delete rsews->m_xyz;
//			m_railStartEditList.removeOne(rsews);
//			delete rsews;
		}
		rsews->m_lineEdit->setVisible(false);
		rsews->m_status->setVisible(false);
		rsews->m_xyz = nullptr;
		qle->setStyleSheet("");
	} else {
		XYZ* xyz = new XYZ();
		bool success;
		xyz->fromString(let, &m_yamceDoc->m_mapSettings.m_railPointSpaces, &success);
		if (!success) {
			qle->setStyleSheet(s_ErrorStyle);
			delete xyz;
			maybeEnableOptions();
			return;
		}
		qle->setStyleSheet("");

		if (rsews->m_xyz) {
			*rsews->m_xyz = *xyz;
			delete xyz;
		} else {
			m_yamceDoc->m_mapSettings.m_railStartPoints.append(xyz);
			rsews->m_xyz = xyz;
		}
		rsews->m_railRootNeedsValidating = true;
	}
	updateRailRootStatus();
	setupRailStartEdits();
	maybeEnableOptions();
}

RailStartEditWithStatus* RailManDialog::findEditStatus(QLineEdit* lineEdit)
{
	for (int i=0; i<m_railStartEditList.size(); i++) {
		if (m_railStartEditList[i]->m_lineEdit == lineEdit)
			return(m_railStartEditList[i]);
	}
	return(nullptr);
}

static inline void updateIcon(RailStartEditWithStatus* rsews)
{
	if (rsews->m_railRootValidated)
		rsews->m_status->setPixmap(QPixmap(":/resources/icons/Check_16x16.png"));
	else
		rsews->m_status->setPixmap(QPixmap(":/resources/icons/Cancel_16x16.png"));

}

void RailManDialog::updateRailRootStatus()
{
	foreach (RailStartEditWithStatus* rsews, m_railStartEditList) {
		if (!rsews->m_railRootNeedsValidating && !rsews->m_lineEdit->text().isEmpty()) {
			updateIcon(rsews);
			continue;
		}
		rsews->m_status->setPixmap(QPixmap());

		if (m_yamceDoc->m_mapSettings.m_railStartPoints.isEmpty()) {
			rsews->m_railRootNeedsValidating = false;
			rsews->m_railRootValidated = false;
			rsews->m_status->setPixmap(QPixmap());
			continue;
		}
		XYZ* xyz = rsews->m_xyz;
		if (!xyz)
			continue;
		RegionFile* rf = m_world->findRegion(blockToRegion(QPoint(xyz->x, xyz->z)));
		if (!rf) {
			rsews->m_railRootNeedsValidating = false;
			rsews->m_railRootValidated = false;
			updateIcon(rsews);
			continue;
		}
		rsews->m_railRootRegionFile = rf;
		if (!rf->isLoaded()) {
			m_mainWindow->getFileMan()->readFile(rsews->m_railRootRegionFile);
			continue;
		} else {
			fileLoaded(rf);
			continue;
		}
	}
}

void RailManDialog::fileLoaded(NBTFile* file)
{
	foreach (RailStartEditWithStatus* rsews, m_railStartEditList) {
		if (file == rsews->m_railRootRegionFile) {
			RegionFile* rf = static_cast<RegionFile*>(file);
			if (!rsews->m_railRootNeedsValidating)
				continue;
			rsews->m_railRootNeedsValidating = false;
			XYZ* xyz = rsews->m_xyz;

			NBTObject* obj = m_railMan->getRail(rf, *xyz);
			if (!obj)
				rsews->m_railRootValidated = false;
			else
				rsews->m_railRootValidated = true;
			updateRailRootStatus();
		}
	}
}

void RailManDialog::onRailManDone()
{
	m_railManIsRunning = false;
	maybeEnableOptions();
	m_railStatsEdit->setHtml(m_railMan->getTextStats());
	if (m_timer) {
		m_timer->stop();
		delete m_timer;
	}
	m_timer = nullptr;
}

void RailManDialog::onColorButtonClicked()
{
	if (!m_mapColorDialog) {
		m_mapColorDialog = new MapColorDialog(m_mainWindow);
		connect(m_mapColorDialog, SIGNAL(rejected()), this, SLOT(onMapColorDialogClosed()));
		connect(m_mapColorDialog, SIGNAL(accepted()), this, SLOT(onMapColorDialogClosed()));
		connect(m_mapColorDialog, SIGNAL(mapColorSelected(int)), this, SLOT(onMapColorChanged(int)));
	}
	m_mapColorDialog->show();
	m_mapColorDialog->raise();
}

void RailManDialog::onMapColorDialogClosed()
{
	m_mapColorDialog = nullptr;
}

void RailManDialog::onMapColorChanged(int mapColor)
{
	qDebug() << "RailManDialog: onMapColorChanged" << mapColor;
	m_yamceDoc->m_railColor = mapColor;
	setColorButtonIcon();
	emit(mapSettingsChanged());
}

