/******************************************************************************
 * regionfile.cpp - Read and write a Minecraft region file
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>
#include <QFile>
#include <QMessageBox>
#include <QtEndian>

#include "xyz.h"
#include "nbtobject.h"
#include "mainwindow.h"
#include "qcompressor.h"
#include "regionfile.h"

const bool DEBUG = false;

// given an x/z of a chunk, get its index into the RegionFile table
inline int getChunkIndex(const QPoint& p) { return((p.x() & 31) + (p.y() & 31) *32);}

inline quint16 nybble4(void* vp, int index) {
	quint8* bp = static_cast<quint8*>(vp);
	return(index%2 == 0 ? bp[index/2]&0x0F : bp[index/2]>>4&0x0F);
}

quint16 bitwidths[] = {0x00,0x01,0x03,0x07, 0x0F,0x1F,0x3F,0x7F,0xFF,0x1FF,0x3FF,0x7FF,0xFFF,0x1FFF,0x3FFF,0x7FFF,0xFFFF};
inline quint16 nybbleX(void* vp, int index, int width) {
	quint64* lp = static_cast<quint64*>(vp);
	int i = index*width/64;
	int shiftBits = (index*width)%64;
	quint16 res;
	if (shiftBits < 64-width)
		res = lp[i] >> shiftBits & bitwidths[width];
	else
		res = ((lp[i] >> shiftBits) + ((lp[i+1] << (64-shiftBits)))) & bitwidths[width];
	return(res);
}

RegionFile::RegionFile()
	: m_locations(nullptr),
	  m_timestamps(nullptr),
	  m_chunks(nullptr)
{
}

RegionFile::~RegionFile()
{
	unload();
}

void RegionFile::unload()
{
	if (DEBUG)
		qDebug() << "RegionFile:unload " << this->m_fileName;
	if (m_locations) {
		delete[] m_locations;
		m_locations = nullptr;
	}
	if (m_timestamps) {
		delete[] m_timestamps;
		m_timestamps = nullptr;
	}
	if (m_chunks) {
		for (int i=0; i<CPF; i++) {
			if (m_chunks[i])
				delete m_chunks[i];
		}
		delete[] m_chunks;
		m_chunks = nullptr;
	}
	super::unload();
}

bool RegionFile::load(ErrorMsgReceiver* receiver)
{
	m_errorMsgReceiver = receiver;
	QString path = QString("%1/%2").arg(m_dirName, m_fileName);
	QFile f(path);
	if (!f.exists()) {
		qWarning() << "Can't find file '" << path << "'";
		return(false);
	}
	qint64 size = f.size();
	bool res = f.open(QIODevice::ReadOnly);
	if (!res) {
		QString s = QString("RegionFile failed to open file %1").arg(m_fileName);
		errorMessage(s);
		return(false);
	}
	m_locations = new quint32[CPF];
	//qDebug() << "m_locations=" << m_locations;
	qint64 readResult = f.read((char*)m_locations, CPF*sizeof(quint32));
	if (readResult != CPF*sizeof (quint32)) {
		QString s = QString("RegionFile locations read failed. Expected 4096 got %1").arg(readResult);
		errorMessage(s);
		return(false);
	}
	m_timestamps = new quint32[CPF];
	readResult = f.read((char*)m_timestamps, CPF*sizeof(quint32));
	if (readResult != CPF*sizeof (quint32)) {
		QString s = QString("RegionFile timestamps read failed. Expected 4096 got %1").arg(readResult);
		errorMessage(s);
		return(false);
	}
	m_chunks = new NBTObject*[CPF];
	memset(m_chunks, 0, CPF*sizeof(NBTObject*));
	qint64 chunkSize = size - (sizeof(quint32)*CPF*2);
	uchar* compressedChunks = new uchar[chunkSize];
	qint64 bytesRead = 0;
	while (bytesRead < size) {
		uchar* p = compressedChunks+bytesRead;
		qint64 r = f.read((char*)p, chunkSize-bytesRead);
		bytesRead += r;
		if (r == 0)
			break;
		if (r == -1) {
			QString s = QString("RegionFile chunk error reading file %1 got %2 bytes expected %3").arg(m_fileName).arg(bytesRead).arg(size);
			errorMessage(s);
			f.close();
			return(false);
		}
	}
	f.close();
	QByteArray ba;
	//QByteArray uncompressedChunk;
	int chunksLoaded = 0;
	for (int i=0; i<CPF; i++) {
		quint32 d = m_locations[i];
		if (d) {
			d = qFromBigEndian(d);
			int sectorCount = d & 0xFF;
			int offset = d >> 8;
			int chunkBegin = (offset - 2) << 12;
			int chunkLength = sectorCount << 12;
			uchar* compressedChunk = compressedChunks + chunkBegin;
			//int expectedDecompressedLength=qFromBigEndian(*(int*)(compressedChunk));	// XXX: Use this to verify data length
			if (DEBUG)
				qDebug() << "RegionFile:load chunk=" << i  << "offset=" << offset << "sectorCount=" << sectorCount << "chunkBegin=" << chunkBegin << "chunkLength=" << chunkLength;
			ba.clear();
			ba.append((char*)compressedChunk, 4);
			ba.append((char*)compressedChunk+5, chunkLength-5);
			//bool res = QCompressor::gzipDecompress(ba, m_data);
			m_data = qUncompress(ba);
			if (m_data.isEmpty()) {
				QString s = QString("Failed to decompress chunk %1 in %2").arg(i).arg(m_fileName);
				errorMessage(s);
				delete[] compressedChunks;
				return(false);
				//continue;
			}
			m_parseOffset = 0;
			m_chunks[i] = parse(nullptr);
			chunksLoaded++;
		}
	}
	if (DEBUG)
		qDebug() << "RegionFile:load loaded " << chunksLoaded << "chunks from" << this->m_fileName;
	m_loaded = true;
	delete[] compressedChunks;
	return(true);
}

bool RegionFile::save(ErrorMsgReceiver*)
{
	qCritical() << "RegionFile::save called";
	return(false);
}

NBTObject*	RegionFile::getBlockPalette(const XYZ& xyz)
{
	QPoint p;
#ifdef _DEBUGx			// sanity check we are in the right file.
	p = blockToRegion(xyz);
	QStringList sl = this->getFileName().split(".");
	if (sl.size() != 4) {
		QString s = QString("RegionFile:getBlockPalette Whacked out filename %1").arg(this->getFileName());
		qCritical() << s;
		return(nullptr);
	}
	if (sl[1].toInt() != (p.x()>>10) || sl[2].toInt() != (p.y()>>10)) {
		QString s = QString("RegionFile:getBlockPalette I'm lost. Expected to be in region %1/%2 and I'm in %3/%4")
				.arg(xyz.x>>10).arg(xyz.z>>10).arg(sl[1], sl[2]);
		qCritical() << s;
		return(nullptr);
	}
#endif
	p = blockToChunk(xyz);
	int chunkIndex = getChunkIndex(p);
	if (!m_locations[chunkIndex])
		return(nullptr);	// chunk doesn't exist
	int yindex = xyz.y>>4;
	NBTObject* chunkRoot = m_chunks[chunkIndex];
	NBTObject* sections = chunkRoot->findChild("*:Level:Sections");
	NList* section = sections->d_list;
	NBTObject* sobj = nullptr;
	for (int l=0; l<section->objcount; l++) {
		sobj = section->d_compound->at(l);
		NBTObject* yobj = sobj->findChild("*:Y");
		if (!yobj) {
			qCritical() << "getBlockPalette: Y not found";
			return(nullptr);
		}
		qint8 yb = yobj->d_byte;
		if (yb < yindex)
			continue;
		if (yb > yindex)
			return(nullptr);	// section doesn't exist
		break;
	}
	int index = ((xyz.y&0x0F) << 8) + ((xyz.z&0x0F) << 4) + (xyz.x &0x0F);
	NBTObject* blocks = sobj->findChild("*:BlockStates");
	if (!blocks)
		return(nullptr);
	int indexWidth = blocks->size / 64;
	int paletteIndex;
	if (indexWidth == 4)
		paletteIndex = nybble4(blocks->d_long_array, index);
	else
		paletteIndex = nybbleX(blocks->d_long_array, index, indexWidth);
	NBTObject* palette = sobj->findChild("*:Palette");
	if (!palette) {
		qCritical() << "RegionFile:getBlockPalette can't find Palette!";
		return(nullptr);
	}
	NList* plist = palette->d_list;
	NBTObject* result = plist->d_compound->at(paletteIndex);
	return(result);
}
