/******************************************************************************
 * treemodel.cpp - Extend QAbstractItemModel to support our tree
 *
 * yamce - Yet Another Minecraft Editor
 * Copyright(c) 2018 - Dick Balaska, and BuckoSoft.
 *
 * yamce is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * yamce is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/
#include <QDebug>

#include "world.h"
#include "nbtobject.h"
//#include "nbtfile.h"
#include "regionfile.h"
#include "treeitem.h"
#include "treemodel.h"

///////////////////////////////////////////////////////////////////////////////
TreeModel::TreeModel(QObject* parent)
	: QAbstractItemModel(parent),
	  m_parent(parent)
{
	QList<QVariant> rootData;
	rootData << "Name" << "Type" << "Value";
	m_rootItem = new TreeItem(rootData, nullptr);
	connect(this, SIGNAL(dataEdited(const QModelIndex&, const QVariant)), parent, SLOT(onDataEdited(const QModelIndex&, const QVariant)));
}

TreeModel::~TreeModel()
{
	delete m_rootItem;
}

QModelIndex TreeModel::index(int row, int column, const QModelIndex &parent)
			const
{
	if (!hasIndex(row, column, parent))
		return QModelIndex();

	TreeItem *parentItem;

	if (!parent.isValid())
		parentItem = m_rootItem;
	else
		parentItem = static_cast<TreeItem*>(parent.internalPointer());

	TreeItem *childItem = parentItem->child(row);
	if (childItem)
		return createIndex(row, column, childItem);
	else
		return QModelIndex();
}

QModelIndex TreeModel::parent(const QModelIndex &index) const
{
	if (!index.isValid())
		return QModelIndex();

	TreeItem *childItem = static_cast<TreeItem*>(index.internalPointer());
	TreeItem *parentItem = childItem->parentItem();

	if (parentItem == m_rootItem)
		return QModelIndex();

	return createIndex(parentItem->row(), 0, parentItem);
}

int TreeModel::rowCount(const QModelIndex &parent) const
{
	TreeItem *parentItem;
	if (parent.column() > 0)
		return 0;

	if (!parent.isValid())
		parentItem = m_rootItem;
	else
		parentItem = static_cast<TreeItem*>(parent.internalPointer());

	return parentItem->childCount();
}

int TreeModel::columnCount(const QModelIndex& ) const
{
//	if (parent.isValid())
//		return static_cast<TreeItem*>(parent.internalPointer())->columnCount();
//	else
		return m_rootItem->columnCount();
}

QVariant TreeModel::data(const QModelIndex& index, int role) const
{
	if (!index.isValid())
		return QVariant();
	TreeItem* item = static_cast<TreeItem*>(index.internalPointer());
	if (role == Qt::EditRole) {
		return item->data(index.column());

	}
	if (role != Qt::DisplayRole)
		return QVariant();


	return item->data(index.column());
}

bool TreeModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
	if (!index.isValid())
		return(false);
	if (role == Qt::EditRole) {
		emit (dataEdited(index, value));
		return(false);
	}
	if (role != Qt::DisplayRole)
		return(false);
	TreeItem* item = static_cast<TreeItem*>(index.internalPointer());
	item->setDataValue(value, index.column());
	emit (dataChanged(index, index));
	return(true);
}

TreeItem* TreeModel::getTreeItem(const QModelIndex& index)
{
	TreeItem* item = static_cast<TreeItem*>(index.internalPointer());
	return(item);
}

// XXX: temp is readonly (?)
Qt::ItemFlags TreeModel::flags(const QModelIndex &index) const
{
	if (!index.isValid())
		return 0;
	Qt::ItemFlags qif = QAbstractItemModel::flags(index);
	if (index.column() == tmType)	// can't edit the type
		return(qif);
	qif |= Qt::ItemIsEditable;
	return qif;
}

QVariant TreeModel::headerData(int section, Qt::Orientation orientation,
							   int role) const
{
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
		return m_rootItem->data(section);

	return QVariant();
}

void TreeModel::replaceRow(QModelIndex& parent, int row, TreeItem* treeItem)
{
	beginInsertRows(parent, row, row);
	TreeItem* item = static_cast<TreeItem*>(parent.internalPointer());
	if (!item)
		item = m_rootItem;
	item->replaceChild(treeItem, row);
	endInsertRows();
}

TreeItem* TreeModel::removeRow(QModelIndex& parent, int row)
{
	//QModelIndex qmi = index(row, 0, parent);
	beginRemoveRows(parent, row, row);
	TreeItem* item = static_cast<TreeItem*>(parent.internalPointer());
	TreeItem* child = item->removeChild(row);
	endRemoveRows();
	NBTObject* pobj = item->getObject();
	if (pobj && pobj->type == TAG_List) {
		//pobj->d_list->objcount--;
		int newCount = pobj->d_list->objcount;
//		QModelIndex columnSibling = parent.siblingAtColumn(2);
		QModelIndex columnSibling = parent.sibling(parent.row(), 2);
		QMap<int, QVariant> data;
		data[Qt::DisplayRole] = QVariant(newCount);
		bool res = this->setItemData(columnSibling, data);
		qDebug() << "removeRow.setItemData returned" << res;
	}

	return(child);
}

void TreeModel::insertRow(QModelIndex &parent, int row, TreeItem* treeItem)
{
	beginInsertRows(parent, row, row);
	TreeItem* tip = static_cast<TreeItem*>(parent.internalPointer());
	tip->insertChild(treeItem, row);
	endInsertRows();
}

TreeItem* TreeModel::addNBTModelData(NBTObject* root, const QString& label, bool noInsert)
{
	if (root && root->type != TAG_Compound) {
		qCritical() << "TreeModel expected root to be a compound";
		return(nullptr);
	}
	QList<QVariant> columnData;
	if (!root) {
		QString s = QString("--%1").arg(label);
		columnData << QVariant(s);
	} else {
		columnData << QVariant(label);
	}
	TreeItem* ti = new TreeItem(columnData, root, m_rootItem);
	if (!noInsert)
		m_rootItem->appendChild(ti);
	if (!root)
		return(ti);
	foreach(NBTObject* nobj, *root->d_compound) {
		setupTree(nobj, ti);
	}
	return(ti);
}

TreeItem* TreeModel::addRegionModelData(RegionFile* file, bool noInsert)
{
	qDebug() << "TreeModel:addRegionFile" << file->getFileName();
	QList<QVariant> columnData;
	columnData << QVariant(file->getFileName());
	TreeItem* ti = new TreeItem(columnData, nullptr, m_rootItem);
	if (!noInsert)
		m_rootItem->appendChild(ti);
	if (file->m_chunks) {
		for (int i=0; i<CPF; i++) {
			if (file->m_chunks[i]) {
				NBTObject* obj = file->m_chunks[i];
				NBTObject* xobj = obj->findChild("*:Level:xPos");
				NBTObject* zobj = obj->findChild("*:Level:zPos");
				QString x = "??";
				QString z = "??";
				if (xobj)
					x = QString::number(xobj->d_int);
				if (zobj)
					z = QString::number(zobj->d_int);
				QString s = QString("%1  %2/%3").arg(i).arg(x, z);
				columnData.clear();
				columnData << QVariant(s);
				TreeItem* tif = new TreeItem(columnData, obj, ti);
				ti->appendChild(tif);
				foreach(NBTObject* nobj, *obj->d_compound) {
					setupTree(nobj, tif);
				}
			}
		}
	}
	return(ti);
}

void TreeModel::setupTree(NBTObject* obj, TreeItem* parent)
{
	QList<QVariant> columnData;
	columnData << obj->name;
	columnData << nbtTypeNames[obj->type];
	switch (obj->type) {
	case TAG_Byte:
		columnData << QString("%1").arg(obj->d_byte);
		break;
	case TAG_Short:
		columnData << QString("%1").arg(obj->d_short);
		break;
	case TAG_Int:
		columnData << QString("%1").arg(obj->d_int);
		break;
	case TAG_Long:
		columnData << QString("%1").arg(obj->d_long);
		break;
	case TAG_Float:
		columnData << QString("%1").arg(obj->d_float);
		break;
	case TAG_Double:
		columnData << QString("%1").arg(obj->d_double);
		break;
	case TAG_Byte_Array:
		columnData << QString("%1").arg(obj->d_byte_array->size());
		break;
	case TAG_Int_Array:
		columnData << QString("%1").arg(obj->size);
		break;
	case TAG_Long_Array:
		columnData << QString("%1").arg(obj->size);
		break;
	case TAG_String:
		columnData << *obj->d_string;
		break;
	case TAG_List: {
		NList* nlist = obj->d_list;
		columnData << nlist->objcount;
		break;
	}
	}
	TreeItem* ti = new TreeItem(columnData, obj, parent);
	parent->appendChild(ti);
	if (obj->type == TAG_Compound) {
		if (obj->d_compound) {
			foreach(NBTObject* nobj, *obj->d_compound) {
				setupTree(nobj, ti);
			}
		}
	}
	if (obj->type == TAG_List) {
		NList* nlist = obj->d_list;
		if (nlist->objtype == TAG_Compound) {
			for (int i=0; i<nlist->objcount; i++) {
				setupTree(nlist->d_compound->at(i), ti);
			}
		} else {
			for (int i=0; i<nlist->objcount; i++) {
				QList<QVariant> cdata;
				cdata << QString("");
				cdata << nbtTypeNames[nlist->objtype];
				switch (nlist->objtype) {
				case TAG_String:
					cdata << nlist->d_string->at(i);
					break;
				case TAG_Byte:
					cdata << QString("%1").arg(nlist->d_byte[i]);
					break;
				case TAG_Short:
					cdata << QString("%1").arg(nlist->d_short[i]);
					break;
				case TAG_Int:
					cdata << QString("%1").arg(nlist->d_int[i]);
					break;
				case TAG_Long:
					cdata << QString("%1").arg(nlist->d_long[i]);
					break;
				case TAG_Float:
					cdata << QString("%1").arg(nlist->d_float[i]);
					break;
				case TAG_Double:
					cdata << QString("%1").arg(nlist->d_double[i]);
					break;

				}
				TreeItem* lti = new TreeItem(cdata, obj, ti);
				ti->appendChild(lti);
			}
		}
	}
}

TreeItem* TreeModel::getTreeItem(const QString& path, TreeItem* node)
{
	QStringList segments = path.split(":");
	int	index;
	QString	segName = segments[0];
	if ((index = segName.indexOf('[')) != -1) {
		QString s = segName.mid(index).left(s.length()-1);	// capture between the []
		segName = segName.left(index);
		index = s.toInt();
	}
	QString nodeName = node->getName();
	if (nodeName == segName) {
		segments.takeFirst();
		if (index != -1)
			node = node->child(index);
		if (segments.isEmpty())
			return(node);
		QString cname = segments.join(":");
		for (TreeItem* ti : node->getChildren()) {
			TreeItem* res = getTreeItem(cname, ti);
			if (res)
				return(res);
		}
		return(nullptr);
	}
	if (segName == "*") {
		segments.takeFirst();
		QString cname = segments.join(":");
		for (TreeItem* ti : node->getChildren()) {
			TreeItem* res = getTreeItem(cname, ti);
			if (res)
				return(res);
		}
	}
	return(nullptr);
}

QModelIndex TreeModel::getIndex(const QString& path)
{
	QStringList segments = path.split(":");
	QModelIndex qmi;
	TreeItem* ti = m_rootItem;
	qmi = index(0,0);
	while (!segments.isEmpty()) {
		int li;
		QString segName = segments[0];
		if ((li = segName.indexOf('[')) != -1) {
			QString s = segName.mid(li).left(s.length()-1);
			segName = segName.left(li);
			li = s.toInt();
		}
		bool found = false;
		int row = 0;
		for (TreeItem* tti : ti->getChildren()) {
			if (tti->getName() == segName) {
				qmi = index(row, 0, qmi);
				ti = ti->child(row);
				found = true;
				segments.takeFirst();
				break;
			}
			row++;
		}
		if (!found) {
			qWarning() << "TreeModel:getIndex failed" << path;
			return(QModelIndex());
		}
		if (li != -1) {
			qmi = index(li, 0, qmi);
			ti = ti->child(li);
		}
	}
	return(qmi);
}

bool TreeModel::isEmpty()
{
	return(m_rootItem->childCount() == 0);
}
